import requests
import numpy as np
import base64
from typing import List

from riotee_gateway.packet_model import PacketApiSend
from riotee_gateway.packet_model import PacketApiReceive


def decode_dev_id(dev_id_b64: str):
    return np.frombuffer(base64.urlsafe_b64decode(dev_id_b64), dtype=np.uint16)[0]


def encode_data(data):
    return str(base64.urlsafe_b64encode(data), "utf-8")


class GatewayClient(object):
    def __init__(self, host: str = "localhost", port: int = 8000):
        self.__url = f"http://{host}:{port}"

    def convert_dev_id(fn_called):
        """Automatically converts dev_id argument to base64"""

        def _convert_dev_id_wrapped(self, dev_id: int | str, *args):
            if dev_id is None:
                return fn_called(self, None, *args)
            if type(dev_id) is str:
                return fn_called(self, dev_id, *args)
            dev_id_b64 = encode_data(np.uint32(dev_id))
            return fn_called(self, dev_id_b64, *args)

        return _convert_dev_id_wrapped

    def get_devices(self) -> List[str]:
        """Reads the list of all devices from which the gateway has received packets."""
        r = requests.get(f"{self.__url}/devices")
        r.raise_for_status()
        return r.json()

    @convert_dev_id
    def send_packet(self, dev_id: int | str, pkt: PacketApiSend):
        r = requests.post(f"{self.__url}/out/{dev_id}", data=pkt.json())
        r.raise_for_status()

    @convert_dev_id
    def send_ascii(self, dev_id: int | str, text: str):
        pkt = PacketApiSend(data=encode_data(bytes(text, encoding="utf-8")))
        self.send_packet(dev_id, pkt)

    @convert_dev_id
    def get_queue_size(self, dev_id: int | str) -> int:
        """Reads the number of packets in the queue for the corresponding device."""
        r = requests.get(f"{self.__url}/in/{dev_id}/size")
        r.raise_for_status()
        return int(r.json())

    @convert_dev_id
    def get_packet(self, dev_id: int | str, pkt_index: int) -> PacketApiReceive:
        """Retrieves a packet from the gateway's fifo queue."""
        r = requests.get(f"{self.__url}/in/{dev_id}/{pkt_index}")
        r.raise_for_status()

        return PacketApiReceive.from_json(r.json())

    @convert_dev_id
    def delete_packet(self, dev_id: int | str, pkt_index: int):
        """Retrieves a packet from the gateway's fifo queue."""
        r = requests.delete(f"{self.__url}/in/{dev_id}/{pkt_index}")
        r.raise_for_status()
        return r.json()

    @convert_dev_id
    def pop(self, dev_id: int | str) -> PacketApiReceive:
        """Pops the first packet from the gateways' fifo queue for the specified device.."""
        try:
            pkt = self.get_packet(dev_id, 0)
        except requests.HTTPError as e:
            if e.response.status_code == 404:
                raise IndexError("No packets in Queue for device")
            raise
        self.delete_packet(dev_id, 0)
        return pkt

    @convert_dev_id
    def pops(self, dev_id: int | str):
        """Pops all packets from the gateway's fifo queue for the specified device."""
        while True:
            try:
                yield self.pop(dev_id)
            except IndexError:
                return

    @convert_dev_id
    def get_packets(self, dev_id: int | str = None) -> List[PacketApiReceive]:
        """Retrieves all packets from all the gateway's fifo queues."""
        if dev_id is None:
            r = requests.get(f"{self.__url}/in/all/all")
        else:
            r = requests.get(f"{self.__url}/in/{dev_id}/all")
        r.raise_for_status()
        return [PacketApiReceive.from_json(json_dict) for json_dict in r.json()]

    @convert_dev_id
    def delete_packets(self, dev_id: int | str = None):
        """Deletes all packets from all the gateway's fifo queues."""
        if dev_id is None:
            r = requests.delete(f"{self.__url}/in/all/all")
        else:
            r = requests.delete(f"{self.__url}/in/{dev_id}/all")
        r.raise_for_status()
        return r.json()



import asyncio
from fastapi import FastAPI
from fastapi import HTTPException
import logging

from riotee_gateway.packet_model import *
from riotee_gateway.transceiver import Transceiver


class PacketDatabase(object):
    """Stores received packets until they are retrieved over the API."""

    def __init__(self) -> None:
        self.__db = dict()

    def add(self, pkt: PacketApiReceive):
        try:
            self.__db[pkt.dev_id].append(pkt)
        except KeyError:
            self.__db[pkt.dev_id] = [pkt]

    def reset(self, dev_id):
        self.__db[dev_id] = list()

    def get_devices(self):
        return list(self.__db.keys())

    def __getitem__(self, dev_id):
        return self.__db[dev_id]


async def receive_loop(tcv: Transceiver, db: PacketDatabase):
    while True:
        pkt = await tcv.read_packet()
        db.add(pkt)
        logging.debug(f"Got packet from {pkt.dev_id} with ID {pkt.pkt_id} @{pkt.timestamp}")


tcv: Transceiver = None
db = PacketDatabase()
app = FastAPI()


@app.get("/")
async def get_root():
    return "Welcome to the Riotee Gateway!"


@app.get("/devices")
async def get_devices():
    return db.get_devices()


@app.get("/in/all/size")
async def get_all_queue_size():
    n_tot = 0
    for dev_id in db.get_devices():
        n_tot += len(db[dev_id])
    return n_tot


@app.get("/in/all/all")
async def get_all_packets():
    pkts = list()
    for dev_id in db.get_devices():
        pkts += db[dev_id]
    return pkts


@app.delete("/in/all/all")
async def delete_all_packets():
    pkts = list()
    for dev_id in db.get_devices():
        db.reset(dev_id)


@app.get("/in/{dev_id}/size")
async def get_queue_size(dev_id: bytes):
    try:
        return len(db[dev_id])
    except KeyError:
        raise HTTPException(status_code=404, detail="Device not found")


@app.get("/in/{dev_id}/all")
async def get_all_dev_packets(dev_id: bytes):
    try:
        return db[dev_id]
    except KeyError:
        raise HTTPException(status_code=404, detail="Device not found")


@app.delete("/in/{dev_id}/all")
async def delete_all_devpackets(dev_id: bytes):
    try:
        db.reset(dev_id)
    except KeyError:
        raise HTTPException(status_code=404, detail="Device not found")


@app.get("/in/{dev_id}/{index}")
async def get_packet(dev_id: bytes, index: int):
    try:
        return db[dev_id][index]
    except KeyError:
        raise HTTPException(status_code=404, detail="Device not found")
    except IndexError:
        raise HTTPException(status_code=404, detail="Packet not found")


@app.delete("/in/{dev_id}/{index}")
async def delete_packet(dev_id: bytes, index: int):
    try:
        del db[dev_id][index]
    except KeyError:
        raise HTTPException(status_code=404, detail="Item not found")
    except IndexError:
        raise HTTPException(status_code=404, detail="Packet not found")


@app.post("/out/{dev_id}")
async def post_packet(dev_id: bytes, packet: PacketApiSend):
    pkt_tcv = PacketTransceiverSend.from_PacketApiSend(packet, dev_id)
    tcv.send_packet(pkt_tcv)
    return packet


@app.on_event("startup")
async def startup_event():
    await tcv.__enter__()
    asyncio.create_task(receive_loop(tcv, db))


@app.on_event("shutdown")
def shutdown_event():
    tcv.__exit__()



import serial_asyncio
from serial.tools import list_ports
from datetime import datetime
import logging

from riotee_gateway.packet_model import PacketApiReceive
from riotee_gateway.packet_model import PacketTransceiverSend


class Transceiver(object):
    """Represents the nRF board that communicates with the devices wirelessly."""

    USB_PID = 0xC8A2
    USB_VID = 0x1209

    @staticmethod
    def find_serial_port() -> str:
        """Finds the serial port name of an attached gateway dongle based on USB IDs."""
        hits = list()
        for port in list_ports.comports():
            if port.vid == Transceiver.USB_VID and port.pid == Transceiver.USB_PID:
                hits.append(port.device)

        if not hits:
            raise Exception("Couldn't find serial port of Riotee Gateway.")
        elif len(hits) == 1:
            logging.info(f"Found serial port at {hits[0]}")
            return hits[0]
        else:
            raise Exception(f"Found multiple potential devices at {' and '.join(hits)}")

    def __init__(self, port: str = None, baudrate: int = 1000000):
        self.__port = port
        self.__baudrate = baudrate

    async def __enter__(self):
        if self.__port is None:
            self.__port = Transceiver.find_serial_port()

        self.__reader, self.__writer = await serial_asyncio.open_serial_connection(
            url=self.__port, baudrate=self.__baudrate
        )
        return self

    def __exit__(self, *args):
        pass

    async def read_packet(self):
        await self.__reader.readuntil(b"[")
        pkt_str = await self.__reader.readuntil(b"]")
        pkt_str = pkt_str[:-1]
        return PacketApiReceive.from_uart(pkt_str, datetime.now())

    def send_packet(self, pkt: PacketTransceiverSend):
        self.__writer.write(pkt.to_uart())
        logging.debug(pkt.to_uart())
